""" Mattermost module """
import argparse
import functools
from typing import Callable, List, Optional, Text, cast

from mattermostdriver import Driver
from pydantic import BaseModel


class Channel(BaseModel):
    """ Mattermost channel model """
    id: Optional[Text]
    type: Optional[Text]
    header: Optional[Text]
    purpose: Optional[Text]
    display_name: Optional[Text]
    name: Text = ""
    mention_count: Optional[int]
    msg_count: Optional[int]
    msg_unread_count: Optional[int]
    update_at: Optional[int]
    last_post_at: Optional[int]
    total_msg_count: Optional[int]
    dirty: Optional[bool]


class Mattermost:
    """ Mattermost helper class """

    def __init__(self, args: argparse.Namespace) -> None:

        self.api = Driver({
            'url': args.server,
            'login_id': args.user,
            'password': args.password,
            'scheme': 'https',
            'port': args.port,
            'basepath': '/api/v4',
            'verify': not args.no_verify,
            'timeout': 30,
            'request_timeout': 30,

        })

        self.api.login()
        self.user = User(**self.api.users.get_user_by_username(args.user))
        self.channels = Channels()

    @functools.lru_cache(128)
    def get_user(self, user_id: Text) -> Text:
        """ Get username from user_id """
        user = self.api.users.get_user(user_id)

        if user:
            return cast(Text, user["username"])
        return "Unknown"

    # Channels is not defined yet, and Channels depends on Mattermost in typing
    # so we need to quote the return definition
    # https://mypy.readthedocs.io/en/latest/kinds_of_types.html#class-name-forward-references
    def init_channels(self) -> 'Channels':
        """ Initialize channels """

        teams = self.api.teams.get_user_teams(self.user.id)
        self.channels.update(self, self.user.id, teams[0]["id"])

        return self.channels

    def init_websocket(self, func: Callable) -> None:
        """ Initialize websocket """

        self.api.init_websocket(func)


class Channels(BaseModel):
    """ Channels model Keeps a list of channels """

    channels: List[Channel] = []

    def update(self, mm: Mattermost, user_id: Text, team_id: Text) -> None:
        """
        Get list of channels for user. We have to subtract msg_count
        from total_msg_count to get unread message count

        https://mattermost.uservoice.com/forums/306457-general/suggestions/38632564-api-add-new-msg-count-to-users-me-teams-team-i
        """
        channel_members = {
            channel["channel_id"]: channel
            for channel
            in mm.api.channels.get_channel_members_for_user(user_id, team_id)
        }

        self.channels = []
        for channel in mm.api.channels.get_channels_for_user(user_id, team_id):
            # Merge results from channel_members_for_user and channels_for_user
            channel.update(channel_members[channel["id"]])
            channel = Channel(**channel)

            if not channel.display_name and "__" in channel.name:
                # for 1-1 chats, the channel name is "<user_id>__<user_id>" where
                # one of the user1 is yours (probably depends on who opened the private chat
                # we need to check which user id is yours, and then lookup the
                # username of the other user
                user1, user2 = channel.name.split("__")

                if user2 == user_id:
                    channel.display_name = mm.get_user(user1)
                else:
                    channel.display_name = mm.get_user(user2)

            channel.msg_unread_count = channel.total_msg_count - channel.msg_count

            self.channels.append(channel)

    def debug(self) -> None:
        """ Debug output of channels """
        for channel in self.channels:
            print(channel)


class User(BaseModel):
    """ User model """
    id: Text
    username: Text
    first_name: Text
    last_name: Text



""" mmblocks """

import argparse
import logging
import logging.handlers
import os
import sys
from logging import error, debug
from typing import Any, List, Optional, Text

import caep
import passpy
import requests
from urllib3.exceptions import InsecureRequestWarning

CONFIG_ID = "mmtools"
CONFIG_NAME = "config"


class NotFound(Exception):
    """Item not Found"""

    def __init__(self, *args: Any) -> None:
        Exception.__init__(self, *args)


def fatal(message: Text, exit_code: int = 1) -> None:
    """ send message to error log and exit """
    error(message)
    sys.exit(exit_code)


def passpy_store(gpgbinary: Optional[Text] = None) -> passpy.store.Store:
    """ passpy store """
    if not gpgbinary:
        gpgbinary = whereis(["gpg2", "gpg"])
        if not gpgbinary:
            raise NotFound("gpg not found")

    return passpy.store.Store(gpg_bin=gpgbinary)


def whereis(filenames: List[Text]) -> Optional[Text]:
    "Locate file"
    if isinstance(filenames, type(str)):
        filenames = [filenames]

    for filename in filenames:
        for path in os.environ["PATH"].split(":"):
            if os.path.isfile("%s/%s" % (path, filename)):
                return os.path.realpath("%s/%s" % (path, filename))
    return None


def gettpassentry(key: Text) -> Any:
    """ Get pass entry from passpy """
    return passpy_store().get_key(key).split("\n")[0]


def setup_logging(
        loglevel: Text = "debug",
        logfile: Optional[Text] = None,
        prefix: Text = "mmtools",
        maxBytes: int = 10000000,  # 10 MB
        backupCount: int = 5) -> None:
    """ Setup loglevel and optional log to file """
    numeric_level = getattr(logging, loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % loglevel)

    datefmt = "%Y-%m-%d %H:%M:%S"
    formatter = "[%(asctime)s] app=" + prefix + " level=%(levelname)s msg=%(message)s"

    if logfile:
        logdir = os.path.dirname(logfile)

        if not os.path.isdir(logdir):
            os.makedirs(logdir)

        # Support strftime in log file names
        handlers = [
            logging.handlers.RotatingFileHandler(
                logfile,
                maxBytes=maxBytes,
                backupCount=backupCount)]

        logging.basicConfig(
            level=numeric_level,
            handlers=handlers,
            format=formatter,
            datefmt=datefmt)
    else:
        logging.basicConfig(
            level=numeric_level,
            stream=sys.stdout,
            format=formatter,
            datefmt=datefmt)


def parseargs(description: Text) -> argparse.ArgumentParser:
    """ Parse arguments """
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description=description
    )

    parser.add_argument("--server", help="Mattermost Server")
    parser.add_argument("--user", help="Mattermost User")
    parser.add_argument("--port", type=int, default=443, help="Mattermost port")
    parser.add_argument("--ignore", help="Regular expression of channels to ignore")
    parser.add_argument("--no-verify", action='store_true', help="SSL verify")
    parser.add_argument("--logfile", help="Log to file")
    parser.add_argument("--loglevel", default="info", help="Log level (default=INFO)")
    parser.add_argument("--password", help="Mattermost password")
    parser.add_argument(
        "--chat-prefix",
        default="🗨️ ",
        help="Prefix to show on statusbar and notification messages")
    parser.add_argument("--team", help="Mattermost team (optional)")
    parser.add_argument("--password-pass-entry",
                        help="pass entry to insert into password")

    return parser


def handle_args(parser: argparse.ArgumentParser, section: Text) -> argparse.Namespace:
    """ Verify default arguments """

    args = caep.handle_args(parser, CONFIG_ID, CONFIG_NAME, section)

    setup_logging(args.loglevel, args.logfile)

    debug(args)

    args.chat_prefix = args.chat_prefix.strip()

    if not args.server:
        fatal("--server not specified")

    if not args.user:
        fatal("--user not specified")

    if args.no_verify:
        requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

    if args.password_pass_entry:
        args.password = gettpassentry(args.password_pass_entry)

    if not args.password:
        fatal("Must specify either --password or --password-pass-entry")

    return args



""" mmtools - watch """

import argparse
import json
import os
import re
import signal
from logging import debug, info
from subprocess import check_output
from typing import Dict, Text

import notify2

from mmtools.mattermost import Mattermost

from . import arguments


def parseargs() -> argparse.Namespace:
    """ Handle arguments """

    parser = arguments.parseargs("mmwatch")

    parser.add_argument(
        "--no-notify", action='store_true', help="Disable notifications")
    parser.add_argument(
        "--pkill", default="i3blocks",
        help="Send SIGUSR1 to process matching value")

    return arguments.handle_args(parser, "mmwatch")


def notify_send(summary: Text, body: Text) -> None:
    """ Send notification message """
    notify2.Notification(summary, body, "notification-message-im").show()


def get_pid(name: Text) -> int:
    """ Get pid from process name """
    return int(check_output(["pidof", "-s", name]))


class EventHandler:
    """ Event Handler """

    def __init__(
            self,
            mm: Mattermost,
            ignore_channels: Text,
            pkill: Text,
            no_notify: bool,
            chat_prefix: Text):

        self.event_map = {
            "posted": self.event_posted,
            "channel_viewed": self.event_channel_viewed,
        }

        self.ignore_channels = ignore_channels
        self.mm = mm
        self.pkill = pkill
        self.no_notify = no_notify
        self.chat_prefix = chat_prefix

    async def event_channel_viewed(self, event: Dict) -> None:
        """ Websocket event handler for channel views """
        data = event.get("data", {})

        channel_id = data.get("channel_id")

        info(json.dumps(data, indent=4, sort_keys=True))

        info(f"channel viewed: {channel_id}")

    async def event_posted(self, event: Dict) -> None:
        """ Websocket event handler for posts """
        data = event["data"]
        post = data.get("post", {})

        # Do not notify on messages sent from myself
        if post.get("user_id") == self.mm.user.id:
            return

        # Do not notify on system join channel/team
        if post.get("type") in ("system_join_channel", "system_join_team"):
            return

        name = data.get("sender_name").rstrip("@")
        channel_name = data.get("channel_name")

        if "__" in channel_name:
            sender_user_id, _ = channel_name.split("__")

            if sender_user_id == self.mm.user.id:
                return

            channel_name = self.mm.get_user(sender_user_id)

        if self.ignore_channels and re.search(self.ignore_channels, channel_name):
            return

        message = post.get("message", "")

        if not self.no_notify:
            notify_send(f"{self.chat_prefix} {channel_name}/{name}", message[:1024])

        if self.pkill:
            pid = get_pid(self.pkill)
            info("kill SIGUSR2 %s (%s)", pid, self.pkill)
            os.kill(pid, signal.SIGUSR2)

    async def event_handler(self, event: Text) -> None:
        """ Websocket event handler """

        debug_event = (
            "status_change",
            "typing",
            "channel_member_updated",
            "user_added",
            None)

        d = json.loads(event)
        if "data" in d:
            for field, value in d["data"].items():
                # Attempt to parse as json in data fields
                if isinstance(field, str):
                    try:
                        d["data"][field] = json.loads(value)
                    except ValueError:
                        pass
                    except TypeError:
                        pass

        event = d.get("event", None)

        if event in self.event_map:
            await self.event_map[event](d)
        elif event in debug_event:
            debug(json.dumps(d, indent=4, sort_keys=True))
        else:
            info(json.dumps(d, indent=4, sort_keys=True))


def main() -> None:
    """ Main module """
    args = parseargs()

    notify2.init('mmtools')

    mm = Mattermost(args)

    handler = EventHandler(mm, args.ignore, args.pkill, args.no_notify, args.chat_prefix)

    mm.init_websocket(handler.event_handler)


if __name__ == "__main__":
    main()



"""status bar for mattermost"""

from os import path

from setuptools import setup

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), "rb") as f:
    long_description = f.read().decode('utf-8')

setup(
    name="mmtools",
    version="0.0.8",
    author="Fredrik Borg",
    zip_safe=True,
    author_email="fredrikb.borg@gmail.com",
    description="mmtools",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="ISC",
    packages=["mmtools"],
    url="https://github.com/frbor/mmtools",
    entry_points={
        'console_scripts': [
            'mmstatus = mmtools.status:main',
            'mmwatch = mmtools.watch:main',
            'mmconfig= mmtools.config:main',
        ]
    },
    python_requires='>=3.6, <4',
    install_requires=[
        'caep',
        'pydantic',
        'mattermostdriver',
        'passpy',
        'notify2',
        'dbus-python',
        'requests'
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "License :: OSI Approved :: ISC License (ISCL)",
    ],
)



""" mmtools - status """

import argparse
import re
import sys
from typing import Text

import requests

from . import arguments
from .mattermost import Channels, Mattermost


def parseargs() -> argparse.Namespace:
    """ Handle arguments """

    parser = arguments.parseargs("mmstatus")
    parser.add_argument(
        "--channel-color",
        default="#00FF00",
        help="Color to use if unread group messages")
    parser.add_argument(
        "--user-color",
        default="#FF4488",
        help="Color to use if unread user messages")

    return arguments.handle_args(parser, "mmstatus")


def i3blocks(
        channels: Channels,
        ignore: Text,
        channel_color: Text,
        user_color: Text,
        chat_prefix: Text) -> None:
    """ Output channel status in i3blocks format """

    private = [f"{channel.display_name}:{channel.msg_unread_count}"
               for channel in channels.channels
               if channel.msg_unread_count and channel.type == "D"
               and not (ignore and re.search(ignore, channel.name))]
    other = [f"{channel.name}:{channel.msg_unread_count}"
             for channel in channels.channels
             if channel.msg_unread_count and channel.type != "D"
             and not (ignore and re.search(ignore, channel.name))]

    out = chat_prefix

    # Join all channels with pipe
    msg = " | ".join(other + private)

    # If we have prefix and output - insert space between prefix and output
    if msg and chat_prefix:
        msg = " " + msg

    print(out + msg)
    print(out + msg)

    if private:
        print(user_color)
    elif other:
        print(channel_color)


def main() -> None:
    """ Main module """

    args = parseargs()

    try:
        mm = Mattermost(args)

        i3blocks(
            mm.init_channels(),
            args.ignore,
            args.channel_color,
            args.user_color,
            args.chat_prefix)
    except requests.exceptions.ConnectionError:
        msg = f"{args.chat_prefix.strip()} Connection error"
        print(f"{msg}\n{msg}\n#FF0000")
        sys.exit(0)


if __name__ == "__main__":
    main()
